@@ -25,15 +25,16 @@ import time
 import binascii
 import hashlib
 import unicodedata
-# import json
+import json
 import getpass
+import warnings
 
 from mnemonic import Mnemonic
 
+from . import messages as proto
 from . import tools
-# from . import mapping
-from . import messages_pb2 as proto
-from . import types_pb2 as types
+from . import mapping
+from .coins import coins_slip44
 from .debuglink import DebugLink
 
 # Python2 vs Python3
@@ -43,6 +44,9 @@ except NameError:
     pass
 
 
+if sys.version_info.major < 3:
+    warnings.warn("Trezorlib will stop supporting Python2 in next versions.", DeprecationWarning)
+
 # try:
 #     from PIL import Image
 #     SCREENSHOT = True
@@ -51,11 +55,6 @@ except NameError:
 
 SCREENSHOT = False
 
-DEFAULT_CURVE = 'secp256k1'
-
-# monkeypatching: text formatting of protobuf messages
-tools.monkeypatch_google_protobuf_text_format()
-
 
 def getch():
     try:
@@ -84,25 +83,20 @@ def getch():
 
 def get_buttonrequest_value(code):
     # Converts integer code to its string representation of ButtonRequestType
-    return [k for k, v in types.ButtonRequestType.items() if v == code][0]
+    return [k for k in dir(proto.ButtonRequestType) if getattr(proto.ButtonRequestType, k) == code][0]
 
 
 def pprint(msg):
     msg_class = msg.__class__.__name__
     msg_size = msg.ByteSize()
-    """
-    msg_ser = msg.SerializeToString()
-    msg_id = mapping.get_type(msg)
-    msg_json = json.dumps(protobuf_json.pb2json(msg))
-    """
-    if isinstance(msg, proto.FirmwareUpload):
+    if isinstance(msg, proto.FirmwareUpload) or isinstance(msg, proto.SelfTest):
         return "<%s> (%d bytes):\n" % (msg_class, msg_size)
     else:
         return "<%s> (%d bytes):\n%s" % (msg_class, msg_size, msg)
 
 
 def log(msg):
-    sys.stderr.write(str(msg))
+    sys.stderr.write(msg)
     sys.stderr.write('\n')
     sys.stderr.flush()
 
@@ -127,7 +121,6 @@ class field(object):
     def __call__(self, f):
         def wrapped_f(*args, **kwargs):
             ret = f(*args, **kwargs)
-            ret.HasField(self.field)
             return getattr(ret, self.field)
         return wrapped_f
 
@@ -143,7 +136,7 @@ class expect(object):
         def wrapped_f(*args, **kwargs):
             ret = f(*args, **kwargs)
             if not isinstance(ret, self.expected):
-                raise Exception("Got %s, expected %s" % (ret.__class__, self.expected))
+                raise RuntimeError("Got %s, expected %s" % (ret.__class__, self.expected))
             return ret
         return wrapped_f
 
@@ -153,8 +146,8 @@ def session(f):
     # with session activation / deactivation
     def wrapped_f(*args, **kwargs):
         client = args[0]
+        client.transport.session_begin()
         try:
-            client.transport.session_begin()
             return f(*args, **kwargs)
         finally:
             client.transport.session_end()
@@ -172,8 +165,7 @@ def normalize_nfc(txt):
             return unicodedata.normalize('NFC', txt.decode('utf-8'))
         if isinstance(txt, str):
             return unicodedata.normalize('NFC', txt)
-
-    raise Exception('unicode/str or bytes/str expected')
+    raise ValueError('unicode/str or bytes/str expected')
 
 
 class BaseClient(object):
@@ -183,13 +175,16 @@ class BaseClient(object):
         self.transport = transport
         super(BaseClient, self).__init__()  # *args, **kwargs)
 
+    def close(self):
+        pass
+
     def cancel(self):
         self.transport.write(proto.Cancel())
 
     @session
     def call_raw(self, msg):
         self.transport.write(msg)
-        return self.transport.read_blocking()
+        return self.transport.read()
 
     @session
     def call(self, msg):
@@ -200,26 +195,27 @@ class BaseClient(object):
         if handler is not None:
             msg = handler(resp)
             if msg is None:
-                raise Exception("Callback %s must return protobuf message, not None" % handler)
+                raise ValueError("Callback %s must return protobuf message, not None" % handler)
             resp = self.call(msg)
 
         return resp
 
     def callback_Failure(self, msg):
-        if msg.code in (types.Failure_PinInvalid,
-                        types.Failure_PinCancelled, types.Failure_PinExpected):
+        if msg.code in (proto.FailureType.PinInvalid,
+                        proto.FailureType.PinCancelled, proto.FailureType.PinExpected):
             raise PinException(msg.code, msg.message)
 
         raise CallException(msg.code, msg.message)
 
-    def close(self):
-        self.transport.close()
+    def register_message(self, msg):
+        '''Allow application to register custom protobuf message type'''
+        mapping.register_message(msg)
 
 
-class DebugWireMixin(object):
+class VerboseWireMixin(object):
     def call_raw(self, msg):
         log("SENDING " + pprint(msg))
-        resp = super(DebugWireMixin, self).call_raw(msg)
+        resp = super(VerboseWireMixin, self).call_raw(msg)
         log("RECEIVED " + pprint(resp))
         return resp
 
@@ -241,7 +237,8 @@ class TextUIMixin(object):
     def callback_RecoveryMatrix(self, msg):
         if self.recovery_matrix_first_pass:
             self.recovery_matrix_first_pass = False
-            log("Use the numeric keypad to describe positions.  For the word list use only left and right keys. The layout is:")
+            log("Use the numeric keypad to describe positions.  For the word list use only left and right keys.")
+            log("Use backspace to correct an entry.  The keypad layout is:")
             log("    7 8 9     7 | 9")
             log("    4 5 6     4 | 6")
             log("    1 2 3     1 | 3")
@@ -254,18 +251,18 @@ class TextUIMixin(object):
                 return proto.WordAck(word='\x08')
 
             # ignore middle column if only 6 keys requested.
-            if (isinstance(msg.type, types.WordRequestType_Matrix6) and character in ('2', '5', '8')):
+            if msg.type == proto.WordRequestType.Matrix6 and character in ('2', '5', '8'):
                 continue
 
-            if (ord(character) >= ord('1') and ord(character) <= ord('9')):
+            if character.isdigit():
                 return proto.WordAck(word=character)
 
     def callback_PinMatrixRequest(self, msg):
-        if msg.type == 1:
+        if msg.type == proto.PinMatrixRequestType.Current:
             desc = 'current PIN'
-        elif msg.type == 2:
+        elif msg.type == proto.PinMatrixRequestType.NewFirst:
             desc = 'new PIN'
-        elif msg.type == 3:
+        elif msg.type == proto.PinMatrixRequestType.NewSecond:
             desc = 'new PIN again'
         else:
             desc = 'PIN'
@@ -276,22 +273,29 @@ class TextUIMixin(object):
         log("    1 2 3")
         log("Please enter %s: " % desc)
         pin = getpass.getpass('')
+        if not pin.isdigit():
+            raise ValueError('Non-numerical PIN provided')
         return proto.PinMatrixAck(pin=pin)
 
     def callback_PassphraseRequest(self, msg):
+        if os.getenv("PASSPHRASE") is not None:
+            log("Passphrase required. Using PASSPHRASE environment variable.")
+            passphrase = Mnemonic.normalize_string(os.getenv("PASSPHRASE"))
+            return proto.PassphraseAck(passphrase=passphrase)
+
         log("Passphrase required: ")
         passphrase = getpass.getpass('')
         log("Confirm your Passphrase: ")
         if passphrase == getpass.getpass(''):
-            passphrase = normalize_nfc(passphrase)
+            passphrase = Mnemonic.normalize_string(passphrase)
             return proto.PassphraseAck(passphrase=passphrase)
         else:
             log("Passphrase did not match! ")
             exit()
 
     def callback_WordRequest(self, msg):
-        if msg.type in (types.WordRequestType_Matrix9,
-                        types.WordRequestType_Matrix6):
+        if msg.type in (proto.WordRequestType.Matrix9,
+                        proto.WordRequestType.Matrix6):
             return self.callback_RecoveryMatrix(msg)
         log("Enter one word of mnemonic: ")
         word = input()
@@ -353,8 +357,8 @@ class DebugLinkMixin(object):
         # return isinstance(value, TypeError)
         # Evaluate missed responses in 'with' statement
         if self.expected_responses is not None and len(self.expected_responses):
-            raise Exception("Some of expected responses didn't come from device: %s" %
-                            [pprint(x) for x in self.expected_responses])
+            raise RuntimeError("Some of expected responses didn't come from device: %s" %
+                               [pprint(x) for x in self.expected_responses])
 
         # Cleanup
         self.expected_responses = None
@@ -362,7 +366,7 @@ class DebugLinkMixin(object):
 
     def set_expected_responses(self, expected):
         if not self.in_with_statement:
-            raise Exception("Must be called inside 'with' statement")
+            raise RuntimeError("Must be called inside 'with' statement")
         self.expected_responses = expected
 
     def setup_debuglink(self, button, pin_correct):
@@ -370,10 +374,10 @@ class DebugLinkMixin(object):
         self.pin_correct = pin_correct
 
     def set_passphrase(self, passphrase):
-        self.passphrase = normalize_nfc(passphrase)
+        self.passphrase = Mnemonic.normalize_string(passphrase)
 
     def set_mnemonic(self, mnemonic):
-        self.mnemonic = normalize_nfc(mnemonic).split(' ')
+        self.mnemonic = Mnemonic.normalize_string(mnemonic).split(' ')
 
     def call_raw(self, msg):
 
@@ -398,17 +402,18 @@ class DebugLinkMixin(object):
             try:
                 expected = self.expected_responses.pop(0)
             except IndexError:
-                raise CallException(types.Failure_UnexpectedMessage,
+                raise CallException(proto.FailureType.UnexpectedMessage,
                                     "Got %s, but no message has been expected" % pprint(msg))
 
             if msg.__class__ != expected.__class__:
-                raise CallException(types.Failure_UnexpectedMessage,
+                raise CallException(proto.FailureType.UnexpectedMessage,
                                     "Expected %s, got %s" % (pprint(expected), pprint(msg)))
 
-            fields = expected.ListFields()  # only filled (including extensions)
-            for field, value in fields:
-                if not msg.HasField(field.name) or getattr(msg, field.name) != value:
-                    raise CallException(types.Failure_UnexpectedMessage,
+            for field, value in expected.__dict__.items():
+                if value is None or value == []:
+                    continue
+                if getattr(msg, field) != value:
+                    raise CallException(proto.FailureType.UnexpectedMessage,
                                         "Expected %s, got %s" % (pprint(expected), pprint(msg)))
 
     def callback_ButtonRequest(self, msg):
@@ -439,7 +444,7 @@ class DebugLinkMixin(object):
         if pos != 0:
             return proto.WordAck(word=self.mnemonic[pos - 1])
 
-        raise Exception("Unexpected call")
+        raise RuntimeError("Unexpected call")
 
 
 class ProtocolMixin(object):
@@ -457,7 +462,7 @@ class ProtocolMixin(object):
     def init_device(self):
         self.features = expect(proto.Features)(self.call)(proto.Initialize())
         if str(self.features.vendor) not in self.VENDORS:
-            raise Exception("Unsupported device")
+            raise RuntimeError("Unsupported device")
 
     def _get_local_entropy(self):
         return os.urandom(32)
@@ -480,20 +485,8 @@ class ProtocolMixin(object):
             n = n[1:]
 
         # coin_name/a/b/c => 44'/SLIP44_constant'/a/b/c
-        coins = {
-            "Bitcoin": 0,
-            "Testnet": 1,
-            "Namecoin": 7,
-            "Litecoin": 2,
-            "Dogecoin": 3,
-            "Dash": 5,
-            "Ether": 60,
-            "EtherClassic": 61,
-            "Zcash": 133,
-            "Decred": 42
-        }
-        if n[0] in coins:
-            n = ["44'", "%d'" % coins[n[0]]] + n[1:]
+        if n[0] in coins_slip44:
+            n = ["44'", "%d'" % coins_slip44[n[0]]] + n[1:]
 
         path = []
         for x in n:
@@ -514,15 +507,13 @@ class ProtocolMixin(object):
         return path
 
     @expect(proto.PublicKey)
-    def get_public_node(self, n, ecdsa_curve_name=DEFAULT_CURVE, show_display=False, coin_name=None):
+    def get_public_node(self, n, ecdsa_curve_name=None, show_display=False, coin_name=None):
         n = self._convert_prime(n)
-        if not ecdsa_curve_name:
-            ecdsa_curve_name = DEFAULT_CURVE
         return self.call(proto.GetPublicKey(address_n=n, ecdsa_curve_name=ecdsa_curve_name, show_display=show_display, coin_name=coin_name))
 
     @field('address')
     @expect(proto.Address)
-    def get_address(self, coin_name, n, show_display=False, multisig=None, script_type=types.SPENDADDRESS):
+    def get_address(self, coin_name, n, show_display=False, multisig=None, script_type=proto.InputScriptType.SPENDADDRESS):
         n = self._convert_prime(n)
         if multisig:
             return self.call(proto.GetAddress(address_n=n, coin_name=coin_name, show_display=show_display, multisig=multisig, script_type=script_type))
@@ -565,13 +556,31 @@ class ProtocolMixin(object):
 
         response = self.call(msg)
 
-        while response.HasField('data_length'):
+        while response.data_length is not None:
             data_length = response.data_length
             data, chunk = data[data_length:], data[:data_length]
             response = self.call(proto.EthereumTxAck(data_chunk=chunk))
 
         return response.signature_v, response.signature_r, response.signature_s
 
+    @expect(proto.EthereumMessageSignature)
+    def ethereum_sign_message(self, n, message):
+        n = self._convert_prime(n)
+        # Convert message to UTF8 NFC (seems to be a bitcoin-qt standard)
+        message = normalize_nfc(message).encode('utf-8')
+        return self.call(proto.EthereumSignMessage(address_n=n, message=message))
+
+    def ethereum_verify_message(self, address, signature, message):
+        # Convert message to UTF8 NFC (seems to be a bitcoin-qt standard)
+        message = normalize_nfc(message).encode('utf-8')
+        try:
+            resp = self.call(proto.EthereumVerifyMessage(address=address, signature=signature, message=message))
+        except CallException as e:
+            resp = e
+        if isinstance(resp, proto.Success):
+            return True
+        return False
+
     @field('entropy')
     @expect(proto.Entropy)
     def get_entropy(self, size):
@@ -606,6 +615,13 @@ class ProtocolMixin(object):
         self.init_device()  # Reload Features
         return out
 
+    @field('message')
+    @expect(proto.Success)
+    def apply_flags(self, flags):
+        out = self.call(proto.ApplyFlags(flags=flags))
+        self.init_device()  # Reload Features
+        return out
+
     @field('message')
     @expect(proto.Success)
     def clear_session(self):
@@ -619,29 +635,164 @@ class ProtocolMixin(object):
         return ret
 
     @expect(proto.MessageSignature)
-    def sign_message(self, coin_name, n, message):
+    def sign_message(self, coin_name, n, message, script_type=proto.InputScriptType.SPENDADDRESS):
         n = self._convert_prime(n)
         # Convert message to UTF8 NFC (seems to be a bitcoin-qt standard)
-        message = normalize_nfc(message).encode("utf-8")
-        return self.call(proto.SignMessage(coin_name=coin_name, address_n=n, message=message))
+        message = normalize_nfc(message).encode('utf-8')
+        return self.call(proto.SignMessage(coin_name=coin_name, address_n=n, message=message, script_type=script_type))
 
     @expect(proto.SignedIdentity)
-    def sign_identity(self, identity, challenge_hidden, challenge_visual, ecdsa_curve_name=DEFAULT_CURVE):
+    def sign_identity(self, identity, challenge_hidden, challenge_visual, ecdsa_curve_name=None):
         return self.call(proto.SignIdentity(identity=identity, challenge_hidden=challenge_hidden, challenge_visual=challenge_visual, ecdsa_curve_name=ecdsa_curve_name))
 
     @expect(proto.ECDHSessionKey)
-    def get_ecdh_session_key(self, identity, peer_public_key, ecdsa_curve_name=DEFAULT_CURVE):
+    def get_ecdh_session_key(self, identity, peer_public_key, ecdsa_curve_name=None):
         return self.call(proto.GetECDHSessionKey(identity=identity, peer_public_key=peer_public_key, ecdsa_curve_name=ecdsa_curve_name))
 
+    @expect(proto.CosiCommitment)
+    def cosi_commit(self, n, data):
+        n = self._convert_prime(n)
+        return self.call(proto.CosiCommit(address_n=n, data=data))
+
+    @expect(proto.CosiSignature)
+    def cosi_sign(self, n, data, global_commitment, global_pubkey):
+        n = self._convert_prime(n)
+        return self.call(proto.CosiSign(address_n=n, data=data, global_commitment=global_commitment, global_pubkey=global_pubkey))
+
     @field('message')
     @expect(proto.Success)
     def set_u2f_counter(self, u2f_counter):
         ret = self.call(proto.SetU2FCounter(u2f_counter=u2f_counter))
         return ret
 
+    @field("address")
+    @expect(proto.NEMAddress)
+    def nem_get_address(self, n, network, show_display=False):
+        n = self._convert_prime(n)
+        return self.call(proto.NEMGetAddress(address_n=n, network=network, show_display=show_display))
+
+    @expect(proto.NEMSignedTx)
+    def nem_sign_tx(self, n, transaction):
+        n = self._convert_prime(n)
+
+        def common_to_proto(common):
+            msg = proto.NEMTransactionCommon()
+            msg.network = (common["version"] >> 24) & 0xFF
+            msg.timestamp = common["timeStamp"]
+            msg.fee = common["fee"]
+            msg.deadline = common["deadline"]
+
+            if "signed" in common:
+                msg.signer = binascii.unhexlify(common["signer"])
+
+            return msg
+
+        def transfer_to_proto(transfer):
+            msg = proto.NEMTransfer()
+            msg.recipient = transfer["recipient"]
+            msg.amount = transfer["amount"]
+
+            if "payload" in transfer["message"]:
+                msg.payload = binascii.unhexlify(transfer["message"]["payload"])
+
+                if transfer["message"]["type"] == 0x02:
+                    msg.public_key = binascii.unhexlify(transfer["message"]["publicKey"])
+
+            if "mosaics" in transfer:
+                msg._extend_mosaics(proto.NEMMosaic(
+                    namespace=mosaic["mosaicId"]["namespaceId"],
+                    mosaic=mosaic["mosaicId"]["name"],
+                    quantity=mosaic["quantity"],
+                ) for mosaic in transfer["mosaics"])
+            return msg
+
+        def aggregate_modification_to_proto(aggregate_modification, msg):
+            msg._extend_modifications(proto.NEMCosignatoryModification(
+                type=modification["modificationType"],
+                public_key=binascii.unhexlify(modification["cosignatoryAccount"]),
+            ) for modification in aggregate_modification["modifications"])
+
+            if "minCosignatories" in aggregate_modification:
+                msg.relative_change = aggregate_modification["minCosignatories"]["relativeChange"]
+
+        def provision_namespace_to_proto(provision_namespace, msg):
+            msg.namespace = provision_namespace["newPart"]
+
+            if provision_namespace["parent"]:
+                msg.parent = provision_namespace["parent"]
+
+            msg.sink = provision_namespace["rentalFeeSink"]
+            msg.fee = provision_namespace["rentalFee"]
+
+        def mosaic_creation_to_proto(mosaic_creation):
+            msg = proto.NEMMosaicCreation()
+            msg.definition.namespace = mosaic_creation["mosaicDefinition"]["id"]["namespaceId"]
+            msg.definition.mosaic = mosaic_creation["mosaicDefinition"]["id"]["name"]
+
+            if mosaic_creation["mosaicDefinition"]["levy"]:
+                msg.definition.levy = mosaic_creation["mosaicDefinition"]["levy"]["type"]
+                msg.definition.fee = mosaic_creation["mosaicDefinition"]["levy"]["fee"]
+                msg.definition.levy_address = mosaic_creation["mosaicDefinition"]["levy"]["recipient"]
+                msg.definition.levy_namespace = mosaic_creation["mosaicDefinition"]["levy"]["mosaicId"]["namespaceId"]
+                msg.definition.levy_mosaic = mosaic_creation["mosaicDefinition"]["levy"]["mosaicId"]["name"]
+
+            msg.definition.description = mosaic_creation["mosaicDefinition"]["description"]
+
+            for property in mosaic_creation["mosaicDefinition"]["properties"]:
+                name = property["name"]
+                value = json.loads(property["value"])
+
+                if name == "divisibility":
+                    msg.definition.divisibility = value
+                elif name == "initialSupply":
+                    msg.definition.supply = value
+                elif name == "supplyMutable":
+                    msg.definition.mutable_supply = value
+                elif name == "transferable":
+                    msg.definition.transferable = value
+
+            msg.sink = mosaic_creation["creationFeeSink"]
+            msg.fee = mosaic_creation["creationFee"]
+            return msg
+
+        def mosaic_supply_change_to_proto(mosaic_supply_change):
+            msg = proto.NEMMosaicSupplyChange()
+            msg.namespace = mosaic_supply_change["mosaicId"]["namespaceId"]
+            msg.mosaic = mosaic_supply_change["mosaicId"]["name"]
+            msg.type = mosaic_supply_change["supplyType"]
+            msg.delta = mosaic_supply_change["delta"]
+            return msg
+
+        msg = proto.NEMSignTx()
+
+        msg.transaction = common_to_proto(transaction)
+        msg.transaction._extend_address_n(n)
+        msg.cosigning = (transaction["type"] == 0x1002)
+
+        if msg.cosigning or transaction["type"] == 0x1004:
+            transaction = transaction["otherTrans"]
+            msg.multisig = common_to_proto(transaction)
+        elif "otherTrans" in transaction:
+            raise CallException("Transaction does not support inner transaction")
+
+        if transaction["type"] == 0x0101:
+            msg.transfer = transfer_to_proto(transaction)
+        elif transaction["type"] == 0x1001:
+            aggregate_modification_to_proto(transaction, msg.aggregate_modification)
+        elif transaction["type"] == 0x2001:
+            provision_namespace_to_proto(transaction, msg.provision_namespace)
+        elif transaction["type"] == 0x4001:
+            msg = mosaic_creation_to_proto(transaction)
+        elif transaction["type"] == 0x4002:
+            msg.mosaic_supply_change = mosaic_supply_change_to_proto(transaction)
+        else:
+            raise CallException("Unknown transaction type")
+
+        return self.call(msg)
+
     def verify_message(self, coin_name, address, signature, message):
         # Convert message to UTF8 NFC (seems to be a bitcoin-qt standard)
-        message = normalize_nfc(message).encode("utf-8")
+        message = normalize_nfc(message).encode('utf-8')
         try:
             resp = self.call(proto.VerifyMessage(address=address, signature=signature, message=message, coin_name=coin_name))
         except CallException as e:
@@ -687,46 +838,14 @@ class ProtocolMixin(object):
                                               ask_on_decrypt=ask_on_decrypt,
                                               iv=iv))
 
-    @field('tx_size')
-    @expect(proto.TxSize)
-    def estimate_tx_size(self, coin_name, inputs, outputs):
-        msg = proto.EstimateTxSize()
-        msg.coin_name = coin_name
-        msg.inputs_count = len(inputs)
-        msg.outputs_count = len(outputs)
-        return self.call(msg)
-
-    def _prepare_simple_sign_tx(self, coin_name, inputs, outputs):
-        msg = proto.SimpleSignTx()
-        msg.coin_name = coin_name
-        msg.inputs.extend(inputs)
-        msg.outputs.extend(outputs)
-
-        known_hashes = []
-        for inp in inputs:
-            if inp.prev_hash in known_hashes:
-                continue
-
-            tx = msg.transactions.add()
-            if self.tx_api:
-                tx.CopyFrom(self.tx_api.get_tx(binascii.hexlify(inp.prev_hash).decode('utf-8')))
-            else:
-                raise Exception('TX_API not defined')
-            known_hashes.append(inp.prev_hash)
-
-        return msg
-
-    def simple_sign_tx(self, coin_name, inputs, outputs):
-        msg = self._prepare_simple_sign_tx(coin_name, inputs, outputs)
-        return self.call(msg).serialized.serialized_tx
-
     def _prepare_sign_tx(self, coin_name, inputs, outputs):
-        tx = types.TransactionType()
-        tx.inputs.extend(inputs)
-        tx.outputs.extend(outputs)
+        tx = proto.TransactionType()
+        tx._extend_inputs(inputs)
+        tx._extend_outputs(outputs)
+        tx._fill_missing()
 
         txes = {}
-        txes[b''] = tx
+        txes[None] = tx
 
         known_hashes = []
         for inp in inputs:
@@ -735,8 +854,9 @@ class ProtocolMixin(object):
 
             if self.tx_api:
                 txes[inp.prev_hash] = self.tx_api.get_tx(binascii.hexlify(inp.prev_hash).decode('utf-8'))
+                txes[inp.prev_hash]._fill_missing()
             else:
-                raise Exception('TX_API not defined')
+                raise RuntimeError('TX_API not defined')
             known_hashes.append(inp.prev_hash)
 
         return txes
@@ -773,24 +893,27 @@ class ProtocolMixin(object):
                 raise CallException("Unexpected message")
 
             # If there's some part of signed transaction, let's add it
-            if res.HasField('serialized') and res.serialized.HasField('serialized_tx'):
-                log("RECEIVED PART OF SERIALIZED TX (%d BYTES)" % len(res.serialized.serialized_tx))
+            if res.serialized and res.serialized.serialized_tx:
+                # log("RECEIVED PART OF SERIALIZED TX (%d BYTES)" % len(res.serialized.serialized_tx))
                 serialized_tx += res.serialized.serialized_tx
 
-            if res.HasField('serialized') and res.serialized.HasField('signature_index'):
+            if res.serialized and res.serialized.signature_index is not None:
                 if signatures[res.serialized.signature_index] is not None:
-                    raise Exception("Signature for index %d already filled" % res.serialized.signature_index)
+                    raise ValueError("Signature for index %d already filled" % res.serialized.signature_index)
                 signatures[res.serialized.signature_index] = res.serialized.signature
 
-            if res.request_type == types.TXFINISHED:
+            if res.request_type == proto.RequestType.TXFINISHED:
                 # Device didn't ask for more information, finish workflow
                 break
 
             # Device asked for one more information, let's process it.
-            current_tx = txes[res.details.tx_hash]
+            if not res.details.tx_hash:
+                current_tx = txes[None]
+            else:
+                current_tx = txes[bytes(res.details.tx_hash)]
 
-            if res.request_type == types.TXMETA:
-                msg = types.TransactionType()
+            if res.request_type == proto.RequestType.TXMETA:
+                msg = proto.TransactionType()
                 msg.version = current_tx.version
                 msg.lock_time = current_tx.lock_time
                 msg.inputs_cnt = len(current_tx.inputs)
@@ -798,22 +921,28 @@ class ProtocolMixin(object):
                     msg.outputs_cnt = len(current_tx.bin_outputs)
                 else:
                     msg.outputs_cnt = len(current_tx.outputs)
-                msg.extra_data_len = len(current_tx.extra_data)
+                msg.extra_data_len = len(current_tx.extra_data) if current_tx.extra_data else 0
                 res = self.call(proto.TxAck(tx=msg))
                 continue
 
-            elif res.request_type == types.TXINPUT:
-                msg = types.TransactionType()
-                msg.inputs.extend([current_tx.inputs[res.details.request_index], ])
+            elif res.request_type == proto.RequestType.TXINPUT:
+                msg = proto.TransactionType()
+                msg._extend_inputs([current_tx.inputs[res.details.request_index], ])
+                if debug_processor is not None:
+                    # If debug_processor function is provided,
+                    # pass thru it the request and prepared response.
+                    # This is useful for unit tests, see test_msg_signtx
+                    msg = debug_processor(res, msg)
+
                 res = self.call(proto.TxAck(tx=msg))
                 continue
 
-            elif res.request_type == types.TXOUTPUT:
-                msg = types.TransactionType()
+            elif res.request_type == proto.RequestType.TXOUTPUT:
+                msg = proto.TransactionType()
                 if res.details.tx_hash:
-                    msg.bin_outputs.extend([current_tx.bin_outputs[res.details.request_index], ])
+                    msg._extend_bin_outputs([current_tx.bin_outputs[res.details.request_index], ])
                 else:
-                    msg.outputs.extend([current_tx.outputs[res.details.request_index], ])
+                    msg._extend_outputs([current_tx.outputs[res.details.request_index], ])
 
                 if debug_processor is not None:
                     # If debug_processor function is provided,
@@ -824,18 +953,18 @@ class ProtocolMixin(object):
                 res = self.call(proto.TxAck(tx=msg))
                 continue
 
-            elif res.request_type == types.TXEXTRADATA:
+            elif res.request_type == proto.RequestType.TXEXTRADATA:
                 o, l = res.details.extra_data_offset, res.details.extra_data_len
-                msg = types.TransactionType()
+                msg = proto.TransactionType()
                 msg.extra_data = current_tx.extra_data[o:o + l]
                 res = self.call(proto.TxAck(tx=msg))
                 continue
 
         if None in signatures:
-            raise Exception("Some signatures are missing!")
+            raise RuntimeError("Some signatures are missing!")
 
-        log("SIGNED IN %.03f SECONDS, CALLED %d MESSAGES, %d BYTES" %
-            (time.time() - start, counter, len(serialized_tx)))
+        # log("SIGNED IN %.03f SECONDS, CALLED %d MESSAGES, %d BYTES" %
+        #    (time.time() - start, counter, len(serialized_tx)))
 
         return (signatures, serialized_tx)
 
@@ -848,12 +977,12 @@ class ProtocolMixin(object):
 
     @field('message')
     @expect(proto.Success)
-    def recovery_device(self, word_count, passphrase_protection, pin_protection, label, language, type=types.RecoveryDeviceType_ScrambledWords, expand=False, dry_run=False):
+    def recovery_device(self, word_count, passphrase_protection, pin_protection, label, language, type=proto.RecoveryDeviceType.ScrambledWords, expand=False, dry_run=False):
         if self.features.initialized and not dry_run:
-            raise Exception("Device is initialized already. Call wipe_device() and try again.")
+            raise RuntimeError("Device is initialized already. Call wipe_device() and try again.")
 
         if word_count not in (12, 18, 24):
-            raise Exception("Invalid word count. Use 12/18/24")
+            raise ValueError("Invalid word count. Use 12/18/24")
 
         self.recovery_matrix_first_pass = True
 
@@ -878,9 +1007,9 @@ class ProtocolMixin(object):
     @field('message')
     @expect(proto.Success)
     @session
-    def reset_device(self, display_random, strength, passphrase_protection, pin_protection, label, language, u2f_counter, skip_backup):
+    def reset_device(self, display_random, strength, passphrase_protection, pin_protection, label, language, u2f_counter=0, skip_backup=False):
         if self.features.initialized:
-            raise Exception("Device is initialized already. Call wipe_device() and try again.")
+            raise RuntimeError("Device is initialized already. Call wipe_device() and try again.")
 
         # Begin with device reset workflow
         msg = proto.ResetDevice(display_random=display_random,
@@ -894,10 +1023,10 @@ class ProtocolMixin(object):
 
         resp = self.call(msg)
         if not isinstance(resp, proto.EntropyRequest):
-            raise Exception("Invalid response, expected EntropyRequest")
+            raise RuntimeError("Invalid response, expected EntropyRequest")
 
         external_entropy = self._get_local_entropy()
-        log("Computer generated entropy: " + binascii.hexlify(external_entropy).decode('ascii'))
+        log("Computer generated entropy: " + binascii.hexlify(external_entropy).decode())
         ret = self.call(proto.EntropyAck(entropy=external_entropy))
         self.init_device()
         return ret
@@ -910,12 +1039,12 @@ class ProtocolMixin(object):
 
     @field('message')
     @expect(proto.Success)
-    def load_device_by_mnemonic(self, mnemonic, pin, passphrase_protection, label, language, skip_checksum=False, expand=False):
+    def load_device_by_mnemonic(self, mnemonic, pin, passphrase_protection, label, language='english', skip_checksum=False, expand=False):
         # Convert mnemonic to UTF8 NKFD
         mnemonic = Mnemonic.normalize_string(mnemonic)
 
         # Convert mnemonic to ASCII stream
-        mnemonic = normalize_nfc(mnemonic)
+        mnemonic = mnemonic.encode('utf-8')
 
         m = Mnemonic('english')
 
@@ -923,10 +1052,10 @@ class ProtocolMixin(object):
             mnemonic = m.expand(mnemonic)
 
         if not skip_checksum and not m.check(mnemonic):
-            raise Exception("Invalid mnemonic checksum")
+            raise ValueError("Invalid mnemonic checksum")
 
         if self.features.initialized:
-            raise Exception("Device is initialized already. Call wipe_device() and try again.")
+            raise RuntimeError("Device is initialized already. Call wipe_device() and try again.")
 
         resp = self.call(proto.LoadDevice(mnemonic=mnemonic, pin=pin,
                                           passphrase_protection=passphrase_protection,
@@ -940,23 +1069,23 @@ class ProtocolMixin(object):
     @expect(proto.Success)
     def load_device_by_xprv(self, xprv, pin, passphrase_protection, label, language):
         if self.features.initialized:
-            raise Exception("Device is initialized already. Call wipe_device() and try again.")
+            raise RuntimeError("Device is initialized already. Call wipe_device() and try again.")
 
         if xprv[0:4] not in ('xprv', 'tprv'):
-            raise Exception("Unknown type of xprv")
+            raise ValueError("Unknown type of xprv")
 
         if len(xprv) < 100 and len(xprv) > 112:
-            raise Exception("Invalid length of xprv")
+            raise ValueError("Invalid length of xprv")
 
-        node = types.HDNodeType()
+        node = proto.HDNodeType()
         data = binascii.hexlify(tools.b58decode(xprv, None))
 
         if data[90:92] != b'00':
-            raise Exception("Contain invalid private key")
+            raise ValueError("Contain invalid private key")
 
         checksum = binascii.hexlify(hashlib.sha256(hashlib.sha256(binascii.unhexlify(data[:156])).digest()).digest()[:4])
         if checksum != data[156:]:
-            raise Exception("Checksum doesn't match")
+            raise ValueError("Checksum doesn't match")
 
         # version 0488ade4
         # depth 00
@@ -983,12 +1112,12 @@ class ProtocolMixin(object):
     @session
     def firmware_update(self, fp):
         if self.features.bootloader_mode is False:
-            raise Exception("Device must be in bootloader mode")
+            raise RuntimeError("Device must be in bootloader mode")
 
         data = fp.read()
 
         resp = self.call(proto.FirmwareErase(length=len(data)))
-        if isinstance(resp, proto.Failure) and resp.code == types.Failure_FirmwareError:
+        if isinstance(resp, proto.Failure) and resp.code == proto.FailureType.FirmwareError:
             return False
 
         # TREZORv1 method
@@ -998,9 +1127,9 @@ class ProtocolMixin(object):
             resp = self.call(proto.FirmwareUpload(payload=data))
             if isinstance(resp, proto.Success):
                 return True
-            elif isinstance(resp, proto.Failure) and resp.code == types.Failure_FirmwareError:
+            elif isinstance(resp, proto.Failure) and resp.code == proto.FailureType.FirmwareError:
                 return False
-            raise Exception("Unexpected result %s" % resp)
+            raise RuntimeError("Unexpected result %s" % resp)
 
         # TREZORv2 method
         if isinstance(resp, proto.FirmwareRequest):
@@ -1013,28 +1142,28 @@ class ProtocolMixin(object):
                     continue
                 elif isinstance(resp, proto.Success):
                     return True
-                elif isinstance(resp, proto.Failure) and resp.code == types.Failure_FirmwareError:
+                elif isinstance(resp, proto.Failure) and resp.code == proto.FailureType.FirmwareError:
                     return False
-                raise Exception("Unexpected result %s" % resp)
+                raise RuntimeError("Unexpected result %s" % resp)
 
-        raise Exception("Unexpected message %s" % resp)
+        raise RuntimeError("Unexpected message %s" % resp)
 
     @field('message')
     @expect(proto.Success)
     def self_test(self):
         if self.features.bootloader_mode is False:
-            raise Exception("Device must be in bootloader mode")
+            raise RuntimeError("Device must be in bootloader mode")
 
-        return self.call(proto.SelfTest())
+        return self.call(proto.SelfTest(payload=b'\x00\xFF\x55\xAA\x66\x99\x33\xCCABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!\x00\xFF\x55\xAA\x66\x99\x33\xCC'))
 
 
 class TrezorClient(ProtocolMixin, TextUIMixin, BaseClient):
     pass
 
 
-class TrezorClientDebug(ProtocolMixin, TextUIMixin, DebugWireMixin, BaseClient):
+class TrezorClientVerbose(ProtocolMixin, TextUIMixin, VerboseWireMixin, BaseClient):
     pass
 
 
-class TrezorDebugClient(ProtocolMixin, DebugLinkMixin, DebugWireMixin, BaseClient):
+class TrezorClientDebugLink(ProtocolMixin, DebugLinkMixin, VerboseWireMixin, BaseClient):
     pass
