@@ -54,7 +54,7 @@ def hash_160_to_bc_address(h160, address_type):
 def compress_pubkey(public_key):
     if byteindex(public_key, 0) == 4:
         return bytes((byteindex(public_key, 64) & 1) + 2) + public_key[1:33]
-    raise Exception("Pubkey is already compressed")
+    raise ValueError("Pubkey is already compressed")
 
 
 def public_key_to_bc_address(public_key, address_type, compress=True):
@@ -120,19 +120,3 @@ def b58decode(v, length):
         return None
 
     return result
-
-
-def monkeypatch_google_protobuf_text_format():
-    # monkeypatching: text formatting of protobuf messages
-    import google.protobuf.text_format
-    import google.protobuf.descriptor
-
-    _oldPrintFieldValue = google.protobuf.text_format.PrintFieldValue
-
-    def _customPrintFieldValue(field, value, out, indent=0, as_utf8=False, as_one_line=False, pointy_brackets=False, float_format=None):
-        if field.type == google.protobuf.descriptor.FieldDescriptor.TYPE_BYTES:
-            _oldPrintFieldValue(field, 'hex(%s)' % binascii.hexlify(value), out, indent, as_utf8, as_one_line)
-        else:
-            _oldPrintFieldValue(field, value, out, indent, as_utf8, as_one_line)
-
-    google.protobuf.text_format.PrintFieldValue = _customPrintFieldValue
